/**
 * Copyright (c) 2008 Washington University
 */
package org.nrg.xnat.client;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URISyntaxException;
import java.net.URL;

import javax.mail.BodyPart;
import javax.mail.MessagingException;
import javax.mail.internet.InternetHeaders;
import javax.mail.internet.MimeBodyPart;
import javax.mail.internet.MimeMultipart;

import org.apache.http.HttpEntity;
import org.apache.http.HttpException;
import org.apache.http.HttpResponse;
import org.apache.http.StatusLine;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpDelete;
import org.apache.http.client.methods.HttpPut;
import org.apache.http.entity.ByteArrayEntity;
import org.apache.http.entity.FileEntity;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.log4j.BasicConfigurator;
import org.dom4j.Document;
import org.dom4j.io.SAXReader;
import org.dom4j.io.XMLWriter;

/**
 * Simple client for POSTing data/metadata to XNAT engine
 * @author Kevin A. Archie <karchie@npg.wustl.edu>
 */
public final class DataClient {
    private static final String HEADER_CONTENT_TYPE = "Content-Type";
    private static final String HEADER_CONTENT_DISPOSITION = "Content-Disposition";
    private static final String MULTIPART_TYPE_FORMAT = "related; type=\"%s\"";
    private static final String CONTENT_TYPE_DATA = "application/octet-stream";
    private static final String CONTENT_TYPE_METADATA = "application/x-xnat-metadata+xml";
    private static final String CONTENT_DISPOSITION_FORMAT_DATA = "inline; filename=\"%s\"";
    private static final String CONTENT_DISPOSITION_METADATA = "attachment";


    private static HttpEntity getEntity(final File data, final Document metadata) throws IOException,MessagingException {
	final HttpEntity entity;
	final int nParts = (null == data ? 0 : 1) + (null == metadata ? 0 : 1);
	switch (nParts) {
	case 0: 
	    entity = new ByteArrayEntity(new byte[0]);
	    break;
	
	case 1:
	    if (null == data) {
		assert null != metadata;
		
		final ByteArrayOutputStream bos = new ByteArrayOutputStream();
		final XMLWriter writer;
		try {
		    writer = new XMLWriter(bos);
		} catch (UnsupportedEncodingException e) {
		    throw new RuntimeException("default encoding is unsupported?", e);
		}
		writer.write(metadata);
		writer.close();
		final ByteArrayEntity be = new ByteArrayEntity(bos.toByteArray());
		be.setContentType(CONTENT_TYPE_METADATA);
		entity = be;
	    } else {
		assert null == metadata;
		entity = new FileEntity(data, CONTENT_TYPE_DATA);
	    }
	    break;
	    
	case 2:
	    assert null != data;
	    
	    final MimeMultipart mmp = new MimeMultipart(String.format(MULTIPART_TYPE_FORMAT,
		    (null == data) ? CONTENT_TYPE_METADATA : CONTENT_TYPE_DATA));

	    final MimeBodyPart part = new MimeBodyPart();
	    part.attachFile(data);
	    System.out.println("uploading " + part.getContentType() + " data " + data.getName());
	    part.setHeader(HEADER_CONTENT_TYPE, part.getContentType());
	    part.setHeader(HEADER_CONTENT_DISPOSITION,
		    String.format(CONTENT_DISPOSITION_FORMAT_DATA, data.getName()));
	    mmp.addBodyPart(part);

	    assert null != metadata;
	    mmp.addBodyPart(metadataBodyPart(metadata));

	    entity = new MimeMultipartEntity(mmp);

	    break;

	default:
	    throw new AssertionError("can't get more than two parts");

	}

	return entity;	
    }


//    public static void doPost(final URL url, final File data, final Document metadata)
//    throws URISyntaxException,IOException,MessagingException,HttpException {
//	final HttpPost post = new HttpPost(url.toURI());
//	final HttpEntity entity = getEntity(data, metadata);
//
//	post.setEntity(entity);
//
//	final HttpClient client = new DefaultHttpClient();
//	final HttpResponse response = client.execute(post);
//	final StatusLine status = response.getStatusLine();
//	System.out.println("POST Response " + status);
//	final HttpEntity re = response.getEntity();
//	re.writeTo(System.out);
//    }
//

    public static void doPut(final URL url, final File data, final Document metadata)
    throws URISyntaxException,IOException,MessagingException,HttpException {
	final HttpPut put = new HttpPut(url.toURI());
	final HttpEntity entity = getEntity(data, metadata);

	put.setEntity(entity);

	final HttpClient client = new DefaultHttpClient();
	final HttpResponse response = client.execute(put);
	final StatusLine status = response.getStatusLine();
	System.out.println("PUT Response " + status);
	final HttpEntity re = response.getEntity();
	if (null != re) {
	    re.writeTo(System.out);
	}
    }



    private static final BodyPart metadataBodyPart(final Document d)
    throws MessagingException,IOException {
	final InternetHeaders headers = new InternetHeaders();
	headers.addHeader(HEADER_CONTENT_TYPE, CONTENT_TYPE_METADATA);
	headers.addHeader(HEADER_CONTENT_DISPOSITION, CONTENT_DISPOSITION_METADATA);
	final ByteArrayOutputStream bos = new ByteArrayOutputStream();
	final XMLWriter writer = new XMLWriter(bos);
	writer.write(d);
	return new MimeBodyPart(headers,bos.toByteArray());
    }


    public static void doDelete(final URL url) throws URISyntaxException,IOException,HttpException {
	final HttpDelete delete = new HttpDelete(url.toURI());
	final HttpClient client = new DefaultHttpClient();
	final HttpResponse response = client.execute(delete);
	final StatusLine status = response.getStatusLine();
	System.out.println("Response " + status);
	final HttpEntity re = response.getEntity();
	re.writeTo(System.out);

    }


    /**
     * @param args
     */
    public static void main(final String[] args) throws Exception {
	if (args.length < 1) {
	    System.err.println("Usage: DataClient <url> -delete \n"
		    + "  OR   DataClient <url> [-single] <data-file> [<metadata-file>]\n"
		    + "  OR   DataClient <url> [-single] [-meta <metadata-file>] [<data-file>]\n");
	    return;
	}
	BasicConfigurator.configure();

	final URL url = new URL(args[0]);
	boolean single_request = false;

	File mdf = null;

	int i;
	for (i = 1; i < args.length && args[i].startsWith("-"); i++) {
	    if ("-delete".equals(args[i])) {
		doDelete(url);
		return;
	    } else if ("-single".equals(args[i])) {
		single_request = true;
	    } else if ("-meta".equals(args[i])) {
		mdf = new File(args[++i]);
	    }
	}

	final File dataf = (i < args.length) ? new File(args[i++]) : null;
	if (i < args.length) {	// TODO: verify we haven't already defined metadata
	    mdf = new File(args[i++]);
	}
	final Document metadata = (null == mdf) ? null : new SAXReader().read(mdf);

	if (single_request) {
	    doPut(url, dataf, metadata);
	} else {
	    doPut(url, null, metadata);
	    doPut(url, dataf, null);
	}
    }

}
